"""Run fetcher pipeline, the application-level service."""

from typing import Optional, TypeAlias

import daiquiri

from dbnomics_fetcher_ops.gitlab import init_gitlab_client
from dbnomics_fetcher_ops.model import FetcherMetadata
from dbnomics_fetcher_ops.utils import without_none_values

__all__ = ["run_fetcher_pipeline"]

logger = daiquiri.getLogger(__name__)


def run_fetcher_pipeline(
    provider_slug: str,
    *,
    debug_gitlab: bool = False,
    dry_run: bool = False,
    fetcher_metadata: FetcherMetadata,
    gitlab_private_token: str,
    json_data_from_git: bool = False,
    json_data_pipeline_id: Optional[int] = None,
    source_data_from_git: bool = False,
    source_data_pipeline_id: Optional[int] = None,
    start_from: Optional[str] = None,
) -> None:
    # Create GitLab client.
    gitlab_url = fetcher_metadata.gitlab.base_url
    gl = init_gitlab_client(gitlab_url, enable_debug=debug_gitlab, private_token=gitlab_private_token)

    project_ref = fetcher_metadata.gitlab.fetcher
    fetcher_group_name, fetcher_project_name = project_ref.expand_group_and_name(provider_slug)
    project_path_with_namespace = f"{fetcher_group_name}/{fetcher_project_name}"

    fetcher_project = gl.projects.get(project_path_with_namespace)

    pipeline_variables = to_key_value_pipeline_variables(
        without_none_values(
            {
                "START_FROM": start_from,
                "SOURCE_DATA_PIPELINE_ID": source_data_pipeline_id,
                "JSON_DATA_PIPELINE_ID": json_data_pipeline_id,
                "SOURCE_DATA_FROM_GIT": source_data_from_git,
                "JSON_DATA_FROM_GIT": json_data_from_git,
            }
        )
    )

    logger.debug(
        "About to run a pipeline for the project %r with variables %r...",
        fetcher_project.path_with_namespace,
        pipeline_variables,
    )

    if not dry_run:
        pipeline = fetcher_project.pipelines.create(
            {"ref": fetcher_project.default_branch, "variables": pipeline_variables}
        )
        logger.info("Created pipeline %r for project %r, see also %s", pipeline, fetcher_project, pipeline.web_url)


PipelineVariableValue: TypeAlias = int | str | bool | None


def to_key_value_pipeline_variable(value: PipelineVariableValue) -> str:
    """Convert a Python value to a string understood by GitLab CI as a pipeline variable value."""
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    if isinstance(value, bool):  # Test bool before int because bool is an int in Python!
        return "1" if value else "0"
    if isinstance(value, int):
        return str(value)
    raise TypeError(value)


def to_key_value_pipeline_variables(d: dict[str, PipelineVariableValue]) -> list[dict[str, str]]:
    """Convert a dict of Python values to a list of GitLab CI pipeline variables."""
    return [{"key": k, "value": to_key_value_pipeline_variable(v)} for k, v in d.items()]



#! /usr/bin/env python3

# dbnomics-fetcher-ops -- Manage DBnomics fetchers
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2020 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-fetcher-ops
#
# dbnomics-fetcher-ops is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-fetcher-ops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""CLI to manage DBnomics fetchers."""


import logging

import daiquiri
import typer
from dotenv import load_dotenv
from xdg import xdg_config_home

from . import app_args, loaders
from .commands.apply import apply_command
from .commands.configure import configure_command
from .commands.list import list_command
from .commands.run import run_command
from .commands.undeploy import undeploy_command

# Do this before calling os.getenv().
config_env_file = xdg_config_home() / "dbnomics" / "dbnomics-fetchers.env"
load_dotenv(config_env_file)
load_dotenv()

logger = daiquiri.getLogger(__name__)

app = typer.Typer()
app.command(name="apply")(apply_command)
app.command(name="configure")(configure_command)
app.command(name="list")(list_command)
app.command(name="run")(run_command)
app.command(name="undeploy")(undeploy_command)


@app.callback(context_settings={"help_option_names": ["-h", "--help"]})
def callback(
    ctx: typer.Context,
    debug: bool = typer.Option(False, "-d", "--debug", envvar="DEBUG", help="Display DEBUG log messages"),
    dry_run: bool = typer.Option(False, "-n", "--dry-run", envvar="DRY_RUN", help="Do not execute actions for real"),
    verbose: bool = typer.Option(False, "-v", "--verbose", envvar="VERBOSE", help="Display INFO log messages"),
    fetchers_yml: str = typer.Option(..., envvar="FETCHERS_YML", help="Path or URL to fetchers.yml"),
    pipelines_yml: str = typer.Option(..., envvar="PIPELINES_YML", help="Path or URL to pipelines.yml"),
):
    """Manage DBnomics fetchers."""
    daiquiri.setup()
    daiquiri.set_default_log_levels(
        [(__package__, logging.DEBUG if debug else logging.INFO if verbose else logging.WARNING)]
    )

    fetcher_metadata = loaders.load_fetchers_yml(fetchers_yml)
    pipelines_config = loaders.load_pipelines_yml(pipelines_yml)

    app_args.app_args = app_args.AppArgs(
        debug=debug,
        dry_run=dry_run,
        fetcher_metadata=fetcher_metadata,
        fetchers_yml=fetchers_yml,
        pipelines_config=pipelines_config,
        pipelines_yml=pipelines_yml,
        verbose=verbose,
    )


def main():
    app()


if __name__ == "__main__":
    main()



# dbnomics-fetcher-ops -- Manage DBnomics fetchers
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2020 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-fetcher-ops
#
# dbnomics-fetcher-ops is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-fetcher-ops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from pathlib import Path

import daiquiri
import requests
import validators
from ruamel.yaml import YAML

from .model import FetcherMetadata, PipelinesConfig

__all__ = ["load_fetchers_yml", "load_pipelines_yml"]

logger = daiquiri.getLogger(__name__)

yaml = YAML(typ="safe")


def load_bytes(source: str) -> bytes:
    """Load ``source`` which can be a file path or an URL."""
    if validators.url(source):
        response = requests.get(source)
        response.raise_for_status()
        return response.content

    # Assume it is a file if it's not an URL.
    return Path(source).read_bytes()


def load_fetchers_yml(fetchers_yml_ref: str) -> FetcherMetadata:
    logger.debug("Loading fetcher metadata from %s...", fetchers_yml_ref)
    fetchers_yml_data = load_yml_file(fetchers_yml_ref)
    return FetcherMetadata.parse_obj(fetchers_yml_data)


def load_pipelines_yml(pipelines_yml_ref: str) -> PipelinesConfig:
    logger.debug("Loading pipelines config from %s...", pipelines_yml_ref)
    pipelines_yml_data = load_yml_file(pipelines_yml_ref)
    return PipelinesConfig.parse_obj(pipelines_yml_data)


def load_yml_file(ref: str):
    text = load_bytes(ref)
    return yaml.load(text)



# dbnomics-fetcher-ops -- Manage DBnomics fetchers
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2020 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-fetcher-ops
#
# dbnomics-fetcher-ops is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-fetcher-ops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from typing import List

import typer

from ..model import PipelineVersionError, known_pipeline_versions, validate_pipeline_version


def list_command(
    slug: bool = typer.Option(False, help="Print fetcher slug instead of code"),
    pipeline_versions: List[str] = typer.Option([], "--pipeline-version", help="Filter fetchers by pipeline version"),
):
    """List fetchers.

    Read fetchers from fetchers.yml.

    Useful to build another script based on the output.
    """
    # Defer import to let the cli.callback function write the variable and avoid importing None.
    from ..app_args import app_args

    assert app_args is not None

    # Validate pipeline versions dynamically because Typer can't take advantage of Literal.
    try:
        for pipeline_version in pipeline_versions:
            validate_pipeline_version(pipeline_version)
    except PipelineVersionError as exc:
        typer.echo(f"Unsupported pipeline version {exc.version!r}, known versions are {known_pipeline_versions!r}")
        raise typer.Abort()

    for fetcher in app_args.fetcher_metadata.fetchers:
        if pipeline_versions and fetcher.pipeline not in pipeline_versions:
            continue
        if slug:
            typer.echo(fetcher.provider_slug)
        else:
            typer.echo(fetcher.provider_code)



# dbnomics-fetcher-ops -- Manage DBnomics fetchers
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2020 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-fetcher-ops
#
# dbnomics-fetcher-ops is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-fetcher-ops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
from pathlib import Path
from typing import Iterator, Optional, Set

import typer

from ..model import FetcherMetadata, PipelinesConfig
from ..utils import iter_child_directories
from .configure import UnsupportedPipelineVersion, configure_fetcher
from .undeploy import undeploy_fetcher

logger = logging.getLogger(__name__)


def apply_command(
    git_repositories_root_dir: Optional[Path] = typer.Option(
        None,
        envvar="GIT_REPOSITORIES_ROOT_DIR",
        help="Directory where the fetcher pipeline keeps clones of the source-data and json-data Git repositories",
    ),
    gitlab_private_token: str = typer.Option(
        ..., envvar="GITLAB_PRIVATE_TOKEN", help="Private access token used to authenticate to GitLab API"
    ),
    debug_gitlab: bool = typer.Option(False, help="Show logging debug messages of Python GitLab"),
    solr_url: Optional[str] = typer.Option(None, envvar="SOLR_URL"),
    workspaces_root_dir: Optional[Path] = typer.Option(
        None,
        envvar="WORKSPACES_ROOT_DIR",
        help="Directory where the fetcher pipeline writes data to, during download and convert jobs",
    ),
):
    """Configure all fetchers listed in fetchers.yml and undeploy the remaining ones."""
    # Defer import to let the cli.callback function write the variable and avoid importing None.
    from ..app_args import app_args

    assert app_args is not None

    fetcher_metadata = app_args.fetcher_metadata
    pipelines_config = app_args.pipelines_config

    logger.info("Configuring fetchers defined in %r...", app_args.fetchers_yml)
    configure_defined_fetchers(fetcher_metadata, pipelines_config, debug_gitlab, gitlab_private_token)

    if workspaces_root_dir is not None:
        logger.info(
            "Undeploying fetchers that are not defined in %r, but did run previously...", app_args.fetchers_yml,
        )
        undeploy_remaining_fetchers(
            fetcher_metadata,
            debug_gitlab,
            gitlab_private_token,
            solr_url,
            git_repositories_root_dir,
            workspaces_root_dir,
        )


def configure_defined_fetchers(
    fetcher_metadata: FetcherMetadata, pipelines_config: PipelinesConfig, debug_gitlab: bool, gitlab_private_token: str,
):
    for fetcher_def in fetcher_metadata.fetchers:
        logger.info("Configuring fetcher %r...", fetcher_def.provider_code)
        pipeline_v5_config = pipelines_config.pipeline_versions.v5
        try:
            configure_fetcher(
                fetcher_def,
                debug_gitlab=debug_gitlab,
                fetcher_metadata=fetcher_metadata,
                gitlab_private_token=gitlab_private_token,
                pipeline_v5_config=pipeline_v5_config,
            )
        except UnsupportedPipelineVersion as exc:
            logger.warning(
                "Fetcher %r uses an unsupported pipeline version %r in fetchers.yml, skipping",
                fetcher_def.provider_slug,
                exc.version,
            )


def undeploy_remaining_fetchers(
    fetcher_metadata: FetcherMetadata,
    debug_gitlab: bool,
    gitlab_private_token: str,
    solr_url: Optional[str],
    git_repositories_root_dir: Optional[Path],
    workspaces_root_dir: Path,
):
    remaining_provider_slugs = get_remaining_provider_slugs(fetcher_metadata, workspaces_root_dir)
    for provider_slug in remaining_provider_slugs:
        undeploy_fetcher(
            provider_slug,
            debug_gitlab=debug_gitlab,
            fetcher_metadata=fetcher_metadata,
            git_repositories_root_dir=git_repositories_root_dir,
            gitlab_private_token=gitlab_private_token,
            solr_url=solr_url,
            workspaces_root_dir=workspaces_root_dir,
        )


def get_remaining_provider_slugs(fetcher_metadata: FetcherMetadata, workspaces_root_dir: Path) -> Set[str]:
    provider_slugs_in_fetchers_yml = {fetcher_def.provider_slug for fetcher_def in fetcher_metadata.fetchers}
    provider_slugs_in_workspace = set(iter_provider_slugs_in_workspace(workspaces_root_dir))
    remaining_provider_slugs = provider_slugs_in_workspace - provider_slugs_in_fetchers_yml
    logger.debug("Found %d remaining fetchers: %r", len(remaining_provider_slugs), remaining_provider_slugs)
    return remaining_provider_slugs


def iter_provider_slugs_in_workspace(workspaces_root_dir: Path) -> Iterator[str]:
    for child_dir in iter_child_directories(workspaces_root_dir):
        yield child_dir.name



"""Run fetcher pipeline, the CLI command."""

from typing import Optional

import humanfriendly
import typer
from typer import Option

from dbnomics_fetcher_ops.services.run_fetcher_pipeline import run_fetcher_pipeline

__all__ = ["run_command"]


def run_command(
    debug_gitlab: bool = typer.Option(False, help="Show logging debug messages of Python GitLab"),
    gitlab_private_token: str = Option(..., envvar="GITLAB_PRIVATE_TOKEN"),
    json_data_from_git: bool = Option(False, envvar="JSON_DATA_FROM_GIT"),
    json_data_pipeline_id: Optional[int] = Option(None, envvar="JSON_DATA_PIPELINE_ID"),
    provider_slug: str = Option(..., envvar="PROVIDER_SLUG"),
    source_data_from_git: bool = Option(False, envvar="SOURCE_DATA_FROM_GIT"),
    source_data_pipeline_id: Optional[int] = Option(None, envvar="SOURCE_DATA_PIPELINE_ID"),
    start_from: Optional[str] = Option(None, envvar="START_FROM"),
):
    """Run a pipeline for a fetcher, with options."""
    from ..app_args import app_args

    assert app_args is not None

    run_fetcher_pipeline(
        provider_slug,
        debug_gitlab=debug_gitlab,
        dry_run=app_args.dry_run,
        fetcher_metadata=app_args.fetcher_metadata,
        gitlab_private_token=gitlab_private_token,
        json_data_from_git=json_data_from_git,
        json_data_pipeline_id=json_data_pipeline_id,
        source_data_from_git=source_data_from_git,
        source_data_pipeline_id=source_data_pipeline_id,
        start_from=start_from,
    )


def parse_time_span(value: str) -> int:
    """Parse the value representing a time span.

    The value must respect the "human friendly" format.
    If it's a bare number (without any unit), it will be interpreted as seconds.

    Cf https://humanfriendly.readthedocs.io/en/latest/api.html#humanfriendly.parse_timespan

    Return the duration in seconds as an integer value.
    """
    try:
        return humanfriendly.parse_timespan(value)
    except humanfriendly.InvalidTimespan:
        doc_url = "https://humanfriendly.readthedocs.io/en/latest/api.html#humanfriendly.parse_timespan"
        typer.echo(f"Invalid value for the --min-duration option: {value!r}, cf {doc_url}")
        raise typer.Abort()



# dbnomics-fetcher-ops -- Manage DBnomics fetchers
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2020 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-fetcher-ops
#
# dbnomics-fetcher-ops is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-fetcher-ops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import base64
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import date, timedelta
from time import sleep
from typing import Optional

import daiquiri
import gitlab
import typer
from gitlab.v4.objects import VISIBILITY_PUBLIC, DeployKey, Project, ProjectPipelineSchedule, User
from gitlab.v4.objects.variables import Variable

from ..app_args import check_provider_slug_is_lowercase, get_fetcher_def_not_found_error_message
from ..gitlab import init_gitlab_client
from ..model import FetcherDef, FetcherDefNotFound, FetcherMetadata, FileDef, PipelineV5Config, ScheduleDef
from ..ssh import generate_ssh_key_pair
from ..utils import find_index

logger = daiquiri.getLogger(__name__)

CI_JOBS_KEY = "CI jobs"
DEFAULT_BRANCH = "master"
SSH_PRIVATE_KEY = "SSH_PRIVATE_KEY"

DBNOMICS_NAMESPACE = "dbnomics"
DBNOMICS_DATA_MODEL = "dbnomics-data-model"
DBNOMICS_SOLR = "dbnomics-solr"


def configure_command(
    provider_slug: str = typer.Argument(..., callback=check_provider_slug_is_lowercase),
    gitlab_private_token: str = typer.Option(
        ..., envvar="GITLAB_PRIVATE_TOKEN", help="Private access token used to authenticate to GitLab API"
    ),
    debug_gitlab: bool = typer.Option(False, help="Show logging debug messages of Python GitLab"),
):
    """Configure a fetcher."""
    # Defer import to let the cli.callback function write the variable and avoid importing None.
    from ..app_args import app_args

    assert app_args is not None

    try:
        fetcher_def = app_args.fetcher_metadata.find_fetcher_def_by_provider_slug(provider_slug)
    except FetcherDefNotFound:
        logger.error(get_fetcher_def_not_found_error_message(provider_slug, app_args.fetchers_yml))
        raise typer.Abort()

    pipeline_v5_config = app_args.pipelines_config.pipeline_versions.v5

    try:
        configure_fetcher(
            fetcher_def,
            debug_gitlab=debug_gitlab,
            fetcher_metadata=app_args.fetcher_metadata,
            gitlab_private_token=gitlab_private_token,
            pipeline_v5_config=pipeline_v5_config,
        )
    except UnsupportedPipelineVersion as exc:
        logger.warning(
            "Fetcher %r uses an unsupported pipeline version %r in fetchers.yml, skipping", provider_slug, exc.version
        )


def configure_fetcher(
    fetcher_def: FetcherDef,
    *,
    debug_gitlab: bool,
    fetcher_metadata: FetcherMetadata,
    gitlab_private_token: str,
    pipeline_v5_config: PipelineV5Config,
):
    """Configure a fetcher.

    Called from apply and configure CLI commands.

    Configures GitLab projects of fetcher, json-data and source-data.

    Supports only pipeline v5.

    Cleanups elements from pipelines v1 or v2.
    """
    if fetcher_def.pipeline != "v5":
        raise UnsupportedPipelineVersion(version=fetcher_def.pipeline)

    # Create GitLab client.
    gitlab_url = fetcher_metadata.gitlab.base_url
    gl = init_gitlab_client(gitlab_url, enable_debug=debug_gitlab, private_token=gitlab_private_token)

    ctx = Context(debug_gitlab=debug_gitlab, gitlab_url=gitlab_url, gl=gl, provider_slug=fetcher_def.provider_slug)

    fetcher_project = init_fetcher_project(gl, fetcher_metadata, ctx)
    source_data_project = load_or_create_source_data_project(gl, fetcher_metadata, ctx)
    json_data_project = load_or_create_json_data_project(gl, fetcher_metadata, ctx)

    # Setup CI conf.
    write_fetcher_project_files(fetcher_project, pipeline_v5_config)
    setup_pipeline_variables(fetcher_project, source_data_project, json_data_project, fetcher_def, fetcher_metadata)
    setup_pipeline_schedules(fetcher_project, fetcher_def, ctx)
    ensure_ssh_key_pairs(fetcher_project, source_data_project, json_data_project, ctx)

    cleanup_elements_from_previous_pipeline_versions(fetcher_project, source_data_project, json_data_project, ctx)


@dataclass
class UnsupportedPipelineVersion(Exception):
    version: str


@dataclass
class TakeOwnershipError(Exception):
    new_owner_name: str
    schedule: ProjectPipelineSchedule

    details: Optional[str] = None

    def __str__(self):
        message = f"User {self.new_owner_name!r} could not take ownership of schedule {self.schedule!r}."
        if self.details is not None:
            message += f" {self.details}"
        return message


@dataclass
class Context:
    debug_gitlab: bool
    gitlab_url: str
    gl: gitlab.Gitlab
    provider_slug: str


def cleanup_elements_from_previous_pipeline_versions(
    fetcher_project: Project,
    source_data_project: Project,
    json_data_project: Project,
    ctx: Context,
) -> None:
    """Cleanup elements from pipelines v1 or v2.

    This includes webhooks of source-data and json-data project.
    """
    gl = ctx.gl
    data_model_project = gl.projects.get(f"{DBNOMICS_NAMESPACE}/{DBNOMICS_DATA_MODEL}")
    solr_project = gl.projects.get(f"{DBNOMICS_NAMESPACE}/{DBNOMICS_SOLR}")

    cleanup_source_data_webhooks_from_previous_pipeline_versions(source_data_project, fetcher_project)
    cleanup_json_data_webhooks_from_previous_pipeline_versions(json_data_project, data_model_project, solr_project)


def cleanup_json_data_webhooks_from_previous_pipeline_versions(
    json_data_project: Project, data_model_project: Project, solr_project: Project
) -> None:
    """Cleanup webhooks of a json-data project defined by pipelines v1 or v2.

    This includes:

    - the webhook that triggers a data model validation,
    - the webhook that triggers a Solr indexation.
    """
    url_prefix_template = "https://git.nomics.world/api/v4/projects/{project_id}/ref/master/trigger/pipeline"
    data_model_url_prefix = url_prefix_template.format(project_id=data_model_project.id)
    solr_url_prefix = url_prefix_template.format(project_id=solr_project.id)

    for hook in json_data_project.hooks.list(as_list=False):
        # Check if hook triggers a data model validation.
        if hook.url.startswith(data_model_url_prefix):
            logger.info(
                "Deleting hook %r that triggers a data model validation in a previous version of the pipeline", hook
            )
            hook.delete()

        # Check if hook triggers an indexation.
        elif hook.url.startswith(solr_url_prefix):
            logger.info("Deleting hook %r that triggers a Solr indexation in a previous version of the pipeline", hook)
            hook.delete()


def cleanup_source_data_webhooks_from_previous_pipeline_versions(
    source_data_project: Project, fetcher_project: Project
) -> None:
    """Cleanup webhooks of a source-data project defined by pipelines v1 or v2.

    This includes the webhook that triggers a convert job.
    """
    url_prefix_template = "https://git.nomics.world/api/v4/projects/{project_id}/ref/master/trigger/pipeline"
    fetcher_url_prefix = url_prefix_template.format(project_id=fetcher_project.id)

    for hook in source_data_project.hooks.list(as_list=False):
        # Check if hook triggers a convert job.
        if hook.url.startswith(fetcher_url_prefix):
            logger.info("Deleting hook %r that triggers a convert script in a previous version of the pipeline", hook)
            hook.delete()


def init_fetcher_project(gl: gitlab.Gitlab, fetcher_metadata: FetcherMetadata, ctx: Context) -> Project:
    """Load or fork or create fetcher project."""
    project_ref = fetcher_metadata.gitlab.fetcher
    fetcher_group_name, fetcher_project_name = project_ref.expand_group_and_name(ctx.provider_slug)
    project_path_with_namespace = f"{fetcher_group_name}/{fetcher_project_name}"
    try:
        return gl.projects.get(project_path_with_namespace)
    except gitlab.exceptions.GitlabGetError as exc:
        if exc.response_code != 404:
            raise

    fork_from_raw = project_ref.fork_from
    if fork_from_raw is not None:
        fork_from = project_ref.expand(fork_from_raw, ctx.provider_slug)
        logger.debug("Fetcher project does not exist, forking from %r...", fork_from)
        try:
            upstream_project = gl.projects.get(fork_from)
        except gitlab.exceptions.GitlabGetError as exc:
            if exc.response_code != 404:
                raise
            logger.error("Could not find project %r, skipping fork and creating new project", fork_from)
        else:
            return fork_project(upstream_project, fetcher_group_name, fetcher_project_name, ctx)

    logger.debug("Fetcher project does not exist, creating...")
    fetcher_description = f"Source code of the fetcher for provider {ctx.provider_slug}"
    return create_project(fetcher_group_name, fetcher_project_name, fetcher_description, ctx)


def load_or_create_source_data_project(gl: gitlab.Gitlab, fetcher_metadata: FetcherMetadata, ctx: Context) -> Project:
    source_data_group_name, source_data_project_name = fetcher_metadata.gitlab.source_data.expand_group_and_name(
        ctx.provider_slug
    )
    try:
        return gl.projects.get(f"{source_data_group_name}/{source_data_project_name}")
    except gitlab.exceptions.GitlabGetError as exc:
        if exc.response_code != 404:
            raise
    logger.debug("Source data project does not exist, creating...")
    source_data_description = f"Source data as downloaded from provider {ctx.provider_slug}"
    return create_project(source_data_group_name, source_data_project_name, source_data_description, ctx)


def load_or_create_json_data_project(gl: gitlab.Gitlab, fetcher_metadata: FetcherMetadata, ctx: Context) -> Project:
    json_data_group_name, json_data_project_name = fetcher_metadata.gitlab.json_data.expand_group_and_name(
        ctx.provider_slug
    )
    try:
        return gl.projects.get(f"{json_data_group_name}/{json_data_project_name}")
    except gitlab.exceptions.GitlabGetError as exc:
        if exc.response_code != 404:
            raise
    logger.debug("JSON data project does not exist, creating...")
    json_data_description = "Data following DBnomics data model, converted from provider data"
    return create_project(json_data_group_name, json_data_project_name, json_data_description, ctx)


def get_deploy_key_title(ctx: Context):
    return f"{ctx.provider_slug} {CI_JOBS_KEY}"


def find_variable_by_name(project: Project, name: str) -> Optional[Variable]:
    try:
        return project.variables.get(name)
    except gitlab.exceptions.GitlabGetError as exc:
        if exc.response_code != 404:
            raise
        return None


def find_deploy_key_by_title(ctx: Context, title: str) -> Optional[DeployKey]:
    for key in ctx.gl.deploykeys.list(as_list=False):
        if key.title == title:
            assert isinstance(key, DeployKey), key
            return key
    return None


def find_project_deploy_key_by_title(project: Project, title: str) -> Optional[DeployKey]:
    for key in project.keys.list(as_list=False):
        if key.title == title:
            return key
    return None


def delete_env_variable(project: Project, name: str, ctx: Context):
    logger.debug("Deleting environment variable %r of %r...", name, project.path_with_namespace)
    try:
        variable = project.variables.get(name)
    except gitlab.exceptions.GitlabGetError as exc:
        if exc.response_code != 404:
            raise
        logger.debug("Environment variable %r was not found in %r", name, project.path_with_namespace)
        return
    variable.delete()
    logger.info("Environment variable %r was deleted from %r", variable, project.path_with_namespace)


def delete_deploy_keys(project: Project, ctx: Context):
    logger.debug("Deleting deploy keys of %r...", project.path_with_namespace)
    for key in project.keys.list(as_list=False):
        if key.title != get_deploy_key_title(ctx):
            logger.warning("%r ignored, title: %r", key, key.title)
            continue
        key.delete()
        logger.info("%r deleted from %r", key, project.path_with_namespace)
    else:
        logger.debug("No deploy key found for %r", project.path_with_namespace)


def ensure_ssh_key_pairs(
    fetcher_project: Project,
    source_data_project: Project,
    json_data_project: Project,
    ctx: Context,
):
    """Checks that the SSH key pairs are configured, otherwise configure them.

    In particular ensure that the fetcher project has a SSH_PRIVATE_KEY masked variable,
    and that source-data and json-data projects have a deploy key.
    """
    ssh_private_key_variable = find_variable_by_name(fetcher_project, SSH_PRIVATE_KEY)
    deploy_key_title = get_deploy_key_title(ctx)
    source_data_deploy_key = find_project_deploy_key_by_title(source_data_project, deploy_key_title)
    json_data_deploy_key = find_project_deploy_key_by_title(json_data_project, deploy_key_title)
    if not ssh_private_key_variable or not source_data_deploy_key or not json_data_deploy_key:
        # Do some cleanup.
        delete_env_variable(fetcher_project, SSH_PRIVATE_KEY, ctx)
        delete_deploy_keys(source_data_project, ctx)
        delete_deploy_keys(json_data_project, ctx)

        # Generate a new SSH key pair, set private key to a fetcher project variable,
        # and create a deploy key from SSH public key used by both source-data
        # and json-data projects.
        ssh_public_key, ssh_private_key = generate_ssh_key_pair(ctx.provider_slug)
        fetcher_project.variables.create({"key": SSH_PRIVATE_KEY, "value": ssh_private_key})
        # Do not display private key value.
        logger.info("%r added to %r", SSH_PRIVATE_KEY, fetcher_project.path_with_namespace)
        deploy_key = source_data_project.keys.create(
            {"title": deploy_key_title, "key": ssh_public_key, "can_push": True}
        )
        logger.info("%r enabled for %r", deploy_key, source_data_project.path_with_namespace)
        json_data_project.keys.enable(deploy_key.id)
        json_data_project.keys.update(deploy_key.id, {"can_push": True})
        logger.info("%r enabled for %r", deploy_key, json_data_project.path_with_namespace)
    else:
        logger.debug("SSH key pair is already configured for this fetcher")


def setup_pipeline_variables(
    fetcher_project: Project,
    source_data_project: Project,
    json_data_project: Project,
    fetcher_def: FetcherDef,
    fetcher_metadata: FetcherMetadata,
):
    """Setup pipeline variables in CI settings.

    Some variables are computed, some other are taken from the fetcher project metadata in fetchers.yml.
    """
    computed_variables = {
        "CATEGORY_TREE_UPDATE_STRATEGY": fetcher_def.category_tree_update_strategy,
        "DATASET_UPDATE_STRATEGY": fetcher_def.dataset_update_strategy,
        "DOWNLOAD_MODE": fetcher_def.download_mode,
        "JSON_DATA_REMOTE_SSH_URL": json_data_project.ssh_url_to_repo,
        "PROVIDER_SLUG": fetcher_def.provider_slug,
        "SOURCE_DATA_REMOTE_SSH_URL": source_data_project.ssh_url_to_repo,
    }
    variables = {**fetcher_metadata.gitlab.fetcher.env, **fetcher_def.env, **computed_variables}
    create_or_update_pipeline_variables(fetcher_project, variables)


def write_fetcher_project_files(fetcher_project: Project, pipeline_v5_config: PipelineV5Config):
    """Write fetcher project files defined in pipeline configuration.

    In particular: .gitlab-ci.yml and Dockerfile.

    If the file exists, follow "if_exists" policy (keep or replace).
    """
    for file_def in pipeline_v5_config.fetcher_project.files:
        write_fetcher_project_file(fetcher_project, file_def)


def write_fetcher_project_file(fetcher_project: Project, file_def: FileDef):
    """Write fetcher project file defined in pipeline configuration.

    If the file exists, follow "if_exists" policy (keep or replace).
    """
    try:
        project_file = fetcher_project.files.get(
            file_path=file_def.path, ref=fetcher_project.default_branch or DEFAULT_BRANCH
        )
    except gitlab.exceptions.GitlabGetError as exc:
        if exc.response_code != 404:
            raise

        # File does not exist, create it

        logger.debug("%r not found, adding it to project %r...", file_def.path, fetcher_project.path_with_namespace)
        project_file = fetcher_project.files.create(
            {
                "file_path": file_def.path,
                "branch": fetcher_project.default_branch or DEFAULT_BRANCH,
                "content": file_def.content,
                "author_email": "dbnomics-fetcher-ops@localhost",
                "author_name": "dbnomics-fetcher-ops",
                "commit_message": f"Add {file_def.path} from template found in pipeline configuration YAML file",
            }
        )
        logger.info("%r file %r created in %r", file_def.path, project_file, fetcher_project.path_with_namespace)
    else:
        # File already exists, apply "if_exists" policy

        if file_def.if_exists == "keep":
            logger.debug(
                "%r file %r already exists in %r, keeping it as-is",
                file_def.path,
                project_file,
                fetcher_project.path_with_namespace,
            )
            return

        elif file_def.if_exists == "replace":
            # Replace file if its content has changed
            # Note: although Git ignores empty commits, GitLab API can create some, so we have to check for changes.

            project_file_content = base64.b64decode(project_file.content).decode()
            if project_file_content.strip() == file_def.content.strip():
                logger.debug(
                    "%r found in project %r and content is up to date, doing nothing",
                    file_def.path,
                    fetcher_project.path_with_namespace,
                )
            else:
                logger.debug(
                    "%r found in project %r and content is not up to date, replacing content...",
                    file_def.path,
                    fetcher_project.path_with_namespace,
                )
                project_file.content = file_def.content
                project_file.save(
                    branch=fetcher_project.default_branch or DEFAULT_BRANCH,
                    commit_message=f"Replace {file_def.path} with template found in pipeline configuration YAML file",
                )
                logger.info(
                    "%r file %r updated in %r", file_def.path, project_file, fetcher_project.path_with_namespace
                )

        else:
            raise ValueError(f"Unsupported value for if_exists: {file_def.if_exists}")


def create_or_update_pipeline_variables(fetcher_project: Project, variables: dict):
    """Create or update pipeline variables for the fetcher project in CI settings."""
    for k, v in variables.items():
        variable = find_variable_by_name(fetcher_project, k)
        if variable is None:
            logger.debug("Variable %r was not found, creating it...", k)
            fetcher_project.variables.create({"key": k, "value": v})
            logger.info("Variable %r created", k)
        else:
            is_dirty = False
            if variable.value != v:
                variable.value = v
                is_dirty = True
            if is_dirty:
                logger.debug("Variable %r was found but is not up to date, updating it...", k)
                variable.save()
                logger.info("Variable %r updated", k)
            else:
                logger.debug("Variable %r was found and is up to date, doing nothing", k)


def create_pipeline_schedule(project: Project, schedule_def: ScheduleDef, ctx: Context) -> ProjectPipelineSchedule:
    schedule = project.pipelineschedules.create(
        {
            "active": True,
            "description": "Run fetcher",
            "ref": project.default_branch or DEFAULT_BRANCH,
            "cron": schedule_def.cron,
            "cron_timezone": schedule_def.timezone,
        }
    )
    take_ownership(schedule, schedule_def.owner, ctx)
    logger.info("%r created for %r with owner %r", schedule, project.path_with_namespace, schedule_def.owner)
    return schedule


def take_ownership(schedule: ProjectPipelineSchedule, owner_name: str, ctx: Context):
    gl = ctx.gl
    owners = gl.users.list(search=owner_name)
    if len(owners) == 0:
        raise TakeOwnershipError(new_owner_name=owner_name, schedule=schedule, details="User not found")

    assert isinstance(owners, list) and len(owners) == 1, owners
    owner = owners[0]
    assert isinstance(owner, User), owner

    with create_impersonation_token(
        owner,
        {
            "name": "dbnomics-fetcher-ops",
            "scopes": ["api"],
            "expires_at": (date.today() + timedelta(days=1)).isoformat(),
        },
    ) as owner_impersonation_token:
        gl2 = init_gitlab_client(
            ctx.gitlab_url, enable_debug=ctx.debug_gitlab, private_token=owner_impersonation_token.token
        )
        fetcher_project2 = gl2.projects.get(schedule.project_id)
        schedule2 = fetcher_project2.pipelineschedules.get(schedule.id)
        try:
            schedule2.take_ownership()
        except gitlab.exceptions.GitlabOwnershipError as exc:
            details = None
            if exc.response_code == 403:
                details = 'Forbidden. Hint: add that user to the members of the group with a "maintainer" role.'
            raise TakeOwnershipError(new_owner_name=owner_name, schedule=schedule, details=details) from exc


@contextmanager
def create_impersonation_token(user: User, token_params):
    impersonation_token = user.impersonationtokens.create(token_params)
    try:
        yield impersonation_token
    finally:
        impersonation_token.delete()


def setup_pipeline_schedules(project: Project, fetcher_def: FetcherDef, ctx: Context):
    """Setup pipeline schedules to match exactly the ones defined in fetcher definition.

    Schedules that are not mentioned in fetcher definition are deleted.
    """
    logger.debug("Configuring pipeline schedules for project %r...", project.path_with_namespace)

    schedule_defs = fetcher_def.schedules.copy()

    for schedule in project.pipelineschedules.list(as_list=False):
        schedule_def_index = find_index(
            lambda schedule_def: schedule_def.cron == schedule.cron and schedule_def.timezone == schedule.cron_timezone,
            schedule_defs,
            default=None,
        )
        if schedule_def_index is None:
            logger.info("Deleting schedule %r that is not mentioned in fetcher definition...", schedule)
            schedule.delete()
        else:
            schedule_def = schedule_defs.pop(schedule_def_index)
            if schedule.owner["username"] != schedule_def.owner:
                logger.info(
                    "Owner of existing schedule %r is different from %r, taking ownership...", schedule, schedule_def
                )
                take_ownership(schedule, schedule_def.owner, ctx)

    for schedule_def in schedule_defs:
        logger.info("Creating schedule from %r...", schedule_def)
        create_pipeline_schedule(project, schedule_def, ctx)


def create_project(group_name: str, project_name: str, description: str, ctx: Context):
    group = load_or_create_group(group_name, ctx)
    project = ctx.gl.projects.create(
        {"name": project_name, "namespace_id": group.id, "description": description, "visibility": VISIBILITY_PUBLIC}
    )
    logger.info("Project created: %r", project.path_with_namespace)
    return project


def create_group(group_name: str, ctx: Context):
    gl = ctx.gl
    parent_group_name, _, group_name = group_name.rpartition("/")
    parent_group_id = None
    if parent_group_name:
        parent_group = gl.groups.get(parent_group_name)
        parent_group_id = parent_group.id
    group = gl.groups.create({"name": group_name, "path": group_name, "parent_id": parent_group_id})
    logger.info("Group created: %r", f"{parent_group_name}/{group.path}")
    return group


def load_or_create_group(group_name: str, ctx: Context):
    try:
        return ctx.gl.groups.get(group_name)
    except gitlab.exceptions.GitlabGetError as exc:
        if exc.response_code != 404:
            raise
    logger.debug("Group %r not found, creating...", group_name)
    return create_group(group_name, ctx)


def fork_project(upstream_project: Project, group_name: str, project_name: str, ctx: Context) -> Project:
    """Fork a project.

    Wait for the fork to really happen, it can take several seconds.
    The more pragmatic way to check if the fork was done is to wait for commits to appear.

    Return a Project instance, not a ProjectFork.
    """
    group = load_or_create_group(group_name, ctx)
    project_fork = upstream_project.forks.create({"namespace_id": group.id, "path": project_name})
    project = ctx.gl.projects.get(project_fork.id)
    commits = []
    while True:
        try:
            commits = project.commits.list(per_page=1)
        except gitlab.exceptions.GitlabListError:
            # GitLab can return a 500 error when fetching commits while forking.
            pass
        if commits:
            break
        sleep(1)
    logger.info("Forked project %r from %r", project.path_with_namespace, upstream_project.path_with_namespace)
    return project



"""Utility functions."""

from pathlib import Path
from typing import Callable, Iterable, Iterator, TypeVar

__all__ = ["find_index", "iter_child_directories", "without_none_values"]

T = TypeVar("T")


def find_index(predicate: Callable[[T], bool], items: Iterable[T], default=None) -> int:
    """Find the index of the first item satisfying the predicate."""
    return next((i for i, item in enumerate(items) if predicate(item)), default)


def iter_child_directories(directory: Path) -> Iterator[Path]:
    """Iterate over child directories of a directory."""
    for child in directory.iterdir():
        if child.is_dir():
            yield child


def without_none_values(d: dict) -> dict:
    """Return a copy of d without None values."""
    return {k: v for k, v in d.items() if v is not None}



# dbnomics-fetcher-ops -- Manage DBnomics fetchers
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2020 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-fetcher-ops
#
# dbnomics-fetcher-ops is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-fetcher-ops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from dataclasses import dataclass
from typing import Optional

import typer

from .model import FetcherMetadata, PipelinesConfig

__all__ = ["app_args", "AppArgs", "check_provider_slug_is_lowercase", "get_fetcher_def_not_found_error_message"]

app_args: Optional["AppArgs"] = None


@dataclass
class AppArgs:
    """Script arguments common to all commands."""

    fetcher_metadata: FetcherMetadata
    fetchers_yml: str
    pipelines_config: PipelinesConfig
    pipelines_yml: str

    debug: bool = False
    dry_run: bool = False
    verbose: bool = False


def check_provider_slug_is_lowercase(value: str):
    if value != value.lower():
        raise typer.BadParameter("Provider slug must be lowercase")
    return value


def get_fetcher_def_not_found_error_message(provider_slug: str, fetchers_yml: str) -> str:
    return f"Could not find fetcher definition for provider {provider_slug!r} in {fetchers_yml!r}"



# dbnomics-fetcher-ops -- Manage DBnomics fetchers
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2020 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-fetcher-ops
#
# dbnomics-fetcher-ops is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-fetcher-ops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from functools import partial
from typing import Dict, List, Literal, Optional, Tuple, Union

from boltons.iterutils import first
from pydantic import BaseModel, Field, HttpUrl, validator

__all__ = [
    "FetcherDef",
    "FetcherDefNotFound",
    "FetcherMetadata",
    "GitLabStructure",
    "known_pipeline_versions",
    "PipelinesConfig",
    "PipelineV5Config",
    "PipelineVersion",
    "PipelineVersionError",
    "PipelineVersions",
    "ProjectRef",
    "ProviderCode",
    "ScheduleDef",
    "UnsupportedModelVersion",
    "validate_pipeline_version",
]

known_pipeline_versions = ["v1", "v2", "v5"]
PipelineVersion = Literal["v1", "v2", "v5"]

UpdateStrategy = Literal["merge", "replace"]


@dataclass
class FetcherDefNotFound(Exception):
    provider_slug: str

    def __str__(self) -> str:
        return f"Could not find fetcher definition for provider {self.provider_slug!r}"


@dataclass
class PipelineVersionError(Exception):
    version: str


@dataclass
class UnsupportedModelVersion(Exception):
    """fetchers.yml version is different from the one expected by the client."""

    expected: int
    found: int


def check_model_version(found: int, *, expected: int):
    if found != expected:
        raise UnsupportedModelVersion(expected=expected, found=found)


def validate_pipeline_version(version: str):
    if version not in known_pipeline_versions:
        raise PipelineVersionError(version=version)


class ScheduleDef(BaseModel):
    cron: str
    owner: str
    timezone: str


LabelValue = Union[str, int, float]

DownloadMode = Literal["full", "incremental"]

ProviderCode = str


class FetcherDef(BaseModel):
    """The definition of a fetcher.

    Pipeline versions are described in [this issue](https://git.nomics.world/dbnomics-fetchers/management/-/issues/948).
    """

    provider_code: ProviderCode
    provider_slug: str
    pipeline: PipelineVersion
    category_tree_update_strategy: UpdateStrategy = "merge"
    dataset_update_strategy: UpdateStrategy = "replace"
    download_mode: DownloadMode = "full"
    env: Dict[str, str] = Field(default_factory=dict)
    labels: Dict[str, LabelValue] = Field(default_factory=dict)
    schedules: List[ScheduleDef] = Field(default_factory=list)


class ProjectRef(BaseModel):
    group: str
    name: str
    env: Dict[str, str] = Field(default_factory=dict)
    fork_from: Optional[str] = None

    def expand(self, value: str, provider_slug: str) -> str:
        """Expand the provider slug in the given value."""
        variables = {"PROVIDER_SLUG": provider_slug}
        for k, v in variables.items():
            value = value.replace("{" + k + "}", v)
        return value

    def expand_group_and_name(self, provider_slug: str) -> Tuple[str, str]:
        """Return a tuple with the group name and project name resolved with the given provider slug."""
        return self.expand(self.group, provider_slug), self.expand(self.name, provider_slug)


class GitLabStructure(BaseModel):
    base_url: HttpUrl
    fetcher: ProjectRef
    json_data: ProjectRef
    source_data: ProjectRef

    @validator("base_url")
    def trim_end_slash(cls, v):
        return v.rstrip("/")

    def get_http_clone_url(self, provider_slug: str, *, project_ref: ProjectRef) -> str:
        group, name = project_ref.expand_group_and_name(provider_slug)
        return f"{self.base_url}/{group}/{name}.git"


class FetcherMetadata(BaseModel):
    version: int

    fetchers: List[FetcherDef]
    gitlab: GitLabStructure

    def find_fetcher_def_by_provider_slug(self, provider_slug: str) -> FetcherDef:
        """Find the FetcherDef item corresponding to the provider slug.

        Raise FetcherDefNotFound if not found.
        """
        for fetcher in self.fetchers:
            if fetcher.provider_slug == provider_slug:
                return fetcher
        raise FetcherDefNotFound(provider_slug=provider_slug)

    validator("version", allow_reuse=True)(partial(check_model_version, expected=2))


class FileDef(BaseModel):
    content: str
    if_exists: Literal["keep", "replace"]
    path: str


class PipelineV5ConfigFetcherProject(BaseModel):
    files: List[FileDef]

    def find_file(self, name: str) -> Optional[str]:
        return first(self.files, key=lambda file: file.name == name, default=None)


class PipelineV5Config(BaseModel):
    fetcher_project: PipelineV5ConfigFetcherProject


class PipelineVersions(BaseModel):
    v5: PipelineV5Config


class PipelinesConfig(BaseModel):
    version: int

    pipeline_versions: PipelineVersions

    validator("version", allow_reuse=True)(partial(check_model_version, expected=1))



# dbnomics-fetcher-ops -- Manage DBnomics fetchers
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2021 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-fetcher-ops
#
# dbnomics-fetcher-ops is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-fetcher-ops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gitlab import Gitlab


def init_gitlab_client(gitlab_url: str, *, private_token: str, enable_debug: bool = False) -> Gitlab:
    gl = Gitlab(gitlab_url, private_token=private_token, api_version=4)
    gl.auth()
    if enable_debug:
        gl.enable_debug()
    return gl



# dbnomics-fetcher-ops -- Manage DBnomics fetchers
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2020 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-fetcher-ops
#
# dbnomics-fetcher-ops is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-fetcher-ops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import gitlab
import typer
from dbnomics_solr import DBnomicsSolrClient
from dbnomics_solr.dbnomics_solr_client import ProviderNotFound
from gitlab.v4.objects import Project

from ..app_args import check_provider_slug_is_lowercase, get_fetcher_def_not_found_error_message
from ..gitlab import init_gitlab_client
from ..model import FetcherDefNotFound, FetcherMetadata

logger = logging.getLogger(__name__)


def undeploy_command(
    provider_slug: str = typer.Argument(..., callback=check_provider_slug_is_lowercase),
    git_repositories_root_dir: Optional[Path] = typer.Option(
        None,
        envvar="GIT_REPOSITORIES_ROOT_DIR",
        help="Directory where the fetcher pipeline keeps clones of the source-data and json-data Git repositories",
    ),
    gitlab_private_token: str = typer.Option(
        ..., envvar="GITLAB_PRIVATE_TOKEN", help="Private access token used to authenticate to GitLab API"
    ),
    debug_gitlab: bool = typer.Option(False, help="Show logging debug messages of Python GitLab"),
    solr_url: Optional[str] = typer.Option(None, envvar="SOLR_URL"),
    workspaces_root_dir: Optional[Path] = typer.Option(
        None,
        envvar="WORKSPACES_ROOT_DIR",
        help="Directory where the fetcher pipeline writes data to, during download and convert jobs",
    ),
):
    """Remove data of this provider on a DBnomics instance.

    This command:

    - disables the schedules of the GitLab project of the fetcher
    - deletes all the Solr documents corresponding to this provider
    - deletes the directories used by the fetcher pipeline (source-data and json-data clones, and workspace)

    As a consequence, the provider won't appear on DBnomics website and API anymore.

    This command does not delete:

    - the GitLab project of the source code of the fetcher
    - the GitLab project of "source-data"
    - the GitLab project of "json-data"
    - the "json-data" directory served by the API (this is done by dbnomics-sync-git)
    """
    # Defer import to let the cli.callback function write the variable and avoid importing None.
    from ..app_args import app_args

    assert app_args is not None

    try:
        undeploy_fetcher(
            provider_slug,
            debug_gitlab=debug_gitlab,
            fetcher_metadata=app_args.fetcher_metadata,
            git_repositories_root_dir=git_repositories_root_dir,
            gitlab_private_token=gitlab_private_token,
            solr_url=solr_url,
            workspaces_root_dir=workspaces_root_dir,
        )
    except FetcherDefNotFound:
        logger.error(get_fetcher_def_not_found_error_message(provider_slug, app_args.fetchers_yml))
        raise typer.Abort()


def undeploy_fetcher(
    provider_slug: str,
    *,
    debug_gitlab: bool,
    fetcher_metadata: FetcherMetadata,
    git_repositories_root_dir: Optional[Path],
    gitlab_private_token: str,
    solr_url: Optional[str],
    workspaces_root_dir: Optional[Path],
):
    # Disable GitLab schedules
    disable_schedules(
        provider_slug,
        debug_gitlab=debug_gitlab,
        fetcher_metadata=fetcher_metadata,
        gitlab_private_token=gitlab_private_token,
    )

    # Delete documents from Solr
    if solr_url is not None:
        delete_provider_from_solr(provider_slug, solr_url=solr_url)

    # Delete directories used by the fetcher pipeline
    delete_pipeline_directories(
        provider_slug, git_repositories_root_dir=git_repositories_root_dir, workspaces_root_dir=workspaces_root_dir
    )


def delete_pipeline_directories(
    provider_slug: str, *, git_repositories_root_dir: Optional[Path], workspaces_root_dir: Optional[Path]
):
    if git_repositories_root_dir is not None:
        git_repositories_dir = git_repositories_root_dir / provider_slug
        delete_pipeline_directory(git_repositories_dir)

    if workspaces_root_dir is not None:
        workspace_dir = workspaces_root_dir / provider_slug
        delete_pipeline_directory(workspace_dir)


def delete_pipeline_directory(path: Path):
    if path.is_dir():
        logger.info("Deleting pipeline directory: %r...", str(path))
        shutil.rmtree(path)


def delete_provider_from_solr(provider_slug: str, *, solr_url: str):
    logger.debug("Deleting documents related to provider %r from Solr...", provider_slug)
    dbnomics_solr_client = DBnomicsSolrClient(solr_url)
    try:
        dbnomics_solr_client.delete_by_provider_slug(provider_slug)
    except ProviderNotFound:
        logger.debug("Provider %r (slug) was not found in Solr index, skipping", provider_slug)


def disable_schedules(
    provider_slug: str, *, gitlab_private_token: str, debug_gitlab: bool, fetcher_metadata: FetcherMetadata,
):
    logger.debug("Disabling schedules of fetcher %r...", provider_slug)

    # Create GitLab client.
    gitlab_url = fetcher_metadata.gitlab.base_url
    gl = init_gitlab_client(gitlab_url, enable_debug=debug_gitlab, private_token=gitlab_private_token)

    try:
        fetcher_project = load_fetcher_project(gl, fetcher_metadata, provider_slug)
    except GitLabProjectNotFound as exc:
        logger.debug(
            "Could not find project %r for fetcher source code of provider %r, skipping disabling schedules",
            exc.project_path_with_namespace,
            provider_slug,
        )
        return

    for schedule in fetcher_project.pipelineschedules.list(as_list=False):
        logger.info("Deleting schedule %r for project %r", schedule, fetcher_project.path_with_namespace)
        schedule.delete()


@dataclass
class GitLabProjectNotFound(Exception):
    project_path_with_namespace: str


def load_fetcher_project(gl: gitlab.Gitlab, fetcher_metadata: FetcherMetadata, provider_slug: str) -> Project:
    fetcher_group_name, fetcher_project_name = fetcher_metadata.gitlab.fetcher.expand_group_and_name(provider_slug)
    project_path_with_namespace = f"{fetcher_group_name}/{fetcher_project_name}"
    try:
        fetcher_project = gl.projects.get(project_path_with_namespace)
    except gitlab.exceptions.GitlabGetError as exc:
        if exc.response_code != 404:
            raise
        raise GitLabProjectNotFound(project_path_with_namespace=project_path_with_namespace) from exc
    return fetcher_project



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dbnomics_fetcher_ops',
 'dbnomics_fetcher_ops.commands',
 'dbnomics_fetcher_ops.services']

package_data = \
{'': ['*']}

install_requires = \
['boltons>=21.0.0,<22.0.0',
 'daiquiri>=3.0.1,<4.0.0',
 'dbnomics-solr>=1.1.2,<2.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'python-gitlab>=2.4.0,<3.0.0',
 'requests>=2.24.0,<3.0.0',
 'ruamel.yaml>=0.16.10,<0.17.0',
 'typer>=0.4.1,<0.5.0',
 'validators>=0.20.0,<0.21.0',
 'xdg>=5.1.1,<6.0.0']

entry_points = \
{'console_scripts': ['dbnomics-fetchers = dbnomics_fetcher_ops.cli:main']}

setup_kwargs = {
    'name': 'dbnomics-fetcher-ops',
    'version': '0.4.11',
    'description': 'Manage DBnomics fetchers',
    'long_description': '# DBnomics fetcher ops\n\nManage DBnomics fetchers.\n\n## Usage\n\n### Install\n\n```bash\npip install dbnomics-fetcher-ops\n```\n\n### Configure a fetcher\n\nConfigure:\n\n- GitLab private token: use `--gitlab-private-token` option or `GITLAB_PRIVATE_TOKEN` environment variable. The private token can be a [personal access token](https://docs.gitlab.com/ee/user/profile/personal_access_tokens.html). It must have the `api` scope.\n\nRun:\n\n```bash\ndbnomics-fetchers -v configure scsmich --dry-run\n# If everything seems OK, remove the --dry-run flag:\ndbnomics-fetchers -v configure scsmich\n```\n\n### List fetchers\n\n```bash\ndbnomics-fetchers -v list\n```\n\n### Run fetcher pipelines\n\n```bash\n# Replace PROVIDER_SLUG by the real value:\ndbnomics-fetchers -v run --provider-slug PROVIDER_SLUG\n\n# To run a pipeline for each fetcher:\ndbnomics-fetchers -v list --slug | xargs -I {} dbnomics-fetchers -v run --provider-slug {}\n```\n\n## Development\n\nThis repository uses [Poetry](https://python-poetry.org/).\n\n```bash\n# git clone repo or fork\ncd dbnomics-fetcher-ops\npoetry install\ncp .env.example .env\n```\n\nRun commands with:\n\n```bash\npoetry run dbnomics-fetchers COMMAND\n```\n\nTo use ipdb:\n\n```bash\npoetry shell\n# Find venv dir with "which python"\nipdb3 /path/to/venv/bin/dbnomics-fetchers ...\n```\n',
    'author': 'Christophe Benz',
    'author_email': 'christophe.benz@nomics.world',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# dbnomics-fetcher-ops -- Manage DBnomics fetchers
# By: Christophe Benz <christophe.benz@cepremap.org>
#
# Copyright (C) 2020 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-fetcher-ops
#
# dbnomics-fetcher-ops is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-fetcher-ops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import subprocess
import tempfile
from pathlib import Path
from typing import Tuple


def generate_ssh_key_pair(provider_slug: str) -> Tuple[str, str]:
    with tempfile.NamedTemporaryFile(prefix="_" + provider_slug) as tmpfile:
        private_key_path = Path(tmpfile.name)
    subprocess.run(
        [
            "ssh-keygen",
            "-f",
            str(private_key_path),
            "-t",
            "rsa",
            "-C",
            f"{provider_slug}-fetcher@localhost",
            "-b",
            "4096",
            "-N",
            "",
        ],
        check=True,
    )
    public_key_path = private_key_path.with_suffix(".pub")
    public_key = public_key_path.read_text()
    public_key_path.unlink()
    private_key = private_key_path.read_text()
    private_key_path.unlink()
    return (public_key, private_key)
